%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Paper: Y.M. Cheung and H. Jia,Categorical-and-Numerical-Attribute
%%%% Data Clustering Based on a Unified Similarity Metric without Knowing 
%%% Cluster Number, Pattern Recognition, vol. 46, no. 8, pp. 2228-2238, 2013.  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%% Code of OCIL algorithm for mixed data clustering %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

close all;
clear all;

%% Input the data set
xx = textread('Heart_disease.txt');  %%% Input data sets:  m=4, 2 classes,
%%%303 instances, 164 for class 1 and 139 for class 2. have 13 features (7
%%%categoric and 6 numeric). Maximum have 4 different categorical values. The last column is class label. 
x = xx(:,1:13); %%% samples
true_class = xx(:,14);  %%% class label
true_class = true_class';
k = 2;  %%% number of clusters
m = 4;  %%% the maximum number of different values can be chosen by each atrribute  
x_categoric = [x(:, 2:3), x(:, 6:7), x(:, 9), x(:, 11), x(:, 13)]; %%% categorical part of the data
x_numeric = [x(:, 1), x(:, 4:5), x(:, 8), x(:, 10), x(:, 12)]; %%%% numerical part of the data

[n, d] = size(x);  %%% n: the number of samples;   d: the number of attributes(dimensionality)
[~, d_c] = size(x_categoric);
[~, d_n] = size(x_numeric);

%% Normalization to the numerical part %%%%%%%%%%%%%%%%%
for i=1:d_n
    x_numeric(:,i) = (x_numeric(:,i)-mean(x_numeric(:,i)))./std(x_numeric(:,i));  
end
%%%%% For Statlog Heart, Heart Disease, and Credit Approval data sets in the
%%%%% paper, it is better to normalize the numerical part first.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Implementation of OCIL algorithm %%%%%
T_total = 100; %%% Execute 100 times to get the average performance 
epoch_record = zeros(1, T_total);
error_record = zeros(1, T_total);
time_record = zeros(1, T_total);

for T_times = 1:T_total
    tic

    %% Calculate the weight of each attribute (feature) %%%%%%
    sample_frequency = zeros(m, d_c);
    for i = 1:n
        for j = 1:d_c
            sample_frequency(x_categoric(i, j), j) = sample_frequency(x_categoric(i, j), j) + 1;
        end
    end

    num_of_diffvalue = zeros(1, d_c);
    for j = 1:d_c
        for i = 1:m
            if sample_frequency(i, j) ~= 0
                num_of_diffvalue(j) = num_of_diffvalue(j) + 1;
            end
        end                               
    end

    value_information = zeros(m, d_c);
    for i = 1:m
        for j = 1:d_c
            if sample_frequency(i, j) == 0
                value_information(i, j) = 0;
            else
                value_information(i, j) = -(sample_frequency(i, j)/n)*(log2(sample_frequency(i, j)/n));
            end
        end
    end
    attribute_entropy = sum(value_information)./num_of_diffvalue;
    feature_weight_categoric = (attribute_entropy./sum(attribute_entropy))*(d_c/(d_c+1)); %%% weight of each categorical attribute
    feature_weight_numeric = 1/(d_c + 1); %%% weight of numerical vector
    feature_weight = [feature_weight_categoric, feature_weight_numeric];
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    %% Randomly initialize the k seed points(one for each class) %%%%%
    miu_cate = zeros(k, d_c);  %%%%initialization
    miu_numer = zeros(k, d_n);

    ran = randperm(n);
    for i = 1:k
        miu_cate(i,:) = x_categoric(ran(i), :);   
        miu_numer(i,:) = x_numeric(ran(i), :);
    end
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    %% Assign the data into the initial k clusters %%%%%
    numeric_sum = miu_numer;
    frequency_in_cluster = zeros(d_c, m, k);  %%%% Count the frequency of each attribute value within the initial clusters
    for c = 1:k
        for j = 1:d_c
            frequency_in_cluster(j,miu_cate(c,j),c) = frequency_in_cluster(j,miu_cate(c,j),c)+1;
        end
    end

    cluster_label = zeros(1, n);
    cluster_count = ones(1, k);
    for i = 1:n
        class_label = OCIL_Mixed_Class_assign(x_categoric(i,:), x_numeric(i,:), frequency_in_cluster, cluster_count, numeric_sum, feature_weight);
        cluster_label(i) = class_label;
        cluster_count(class_label) = cluster_count(class_label)+1;
        numeric_sum(class_label, :) = numeric_sum(class_label, :) + x_numeric(i, :); 
        for j = 1:d_c
            frequency_in_cluster(j, x_categoric(i,j), class_label) = frequency_in_cluster(j, x_categoric(i,j), class_label)+1;
        end
    end
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    %% Iterative steps: adjust the cluster label of each data untill no data need to change class. %%%%%%%%%%
    moves = 1;
    epoch = 0;
    while (moves ~= 0) %%% Some class label has been changed in the last step%%%
        moves = 0;
        for i = 1:n
        
            new_label = OCIL_Mixed_Class_assign(x_categoric(i,:), x_numeric(i,:), frequency_in_cluster, cluster_count, numeric_sum, feature_weight);

            if cluster_label(i) ~= new_label;  %%% need adjustment
                moves = moves + 1;
                old_cluster = cluster_label(i);
                cluster_label(i) = new_label;     %%%% updata cluster label
                cluster_count(new_label) = cluster_count(new_label) + 1;
                cluster_count(old_cluster) = cluster_count(old_cluster) - 1;   %%%% update cluster counter
            
                numeric_sum(new_label, :) = numeric_sum(new_label, :) + x_numeric(i, :);
                numeric_sum(old_cluster, :) = numeric_sum(old_cluster, :) - x_numeric(i, :);
            
                for j = 1:d_c        %%%% update the information of the two clusters
                    frequency_in_cluster(j, x_categoric(i,j), new_label) = frequency_in_cluster(j, x_categoric(i,j), new_label) + 1;       
                    frequency_in_cluster(j, x_categoric(i,j), old_cluster) = frequency_in_cluster(j, x_categoric(i,j), old_cluster) - 1; 
                end
            end
        end
        epoch = epoch + 1;
    end
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    %% Record the clustering result %%%%%%
    exetime = toc;
    accuracy = calAccuracy(cluster_label',true_class'); %%% calculate clustering accuracy
    error = 1 - accuracy;

    epoch_record(T_times) = epoch;
    error_record(T_times) = error;
    time_record(T_times) = exetime;

end

%% Summarize the clustering results %%%%%%%%%
error_mean = mean(error_record)
error_std = std(error_record)

epoch_mean = mean(epoch_record)
epoch_std = std(epoch_record)

time_mean = mean(time_record)
time_std = std(time_record)

