%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% Paper: Y.M. Cheung and H. Jia,Categorical-and-Numerical-Attribute
%%%% Data Clustering Based on a Unified Similarity Metric without Knowing 
%%% Cluster Number, Pattern Recognition, vol. 46, no. 8, pp. 2228-2238, 2013.  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Code of PCL-OC algorithm for mixed data clustering without knowing true cluster number %%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

close all;
clear all;

%% Input the data set
xx = textread('Heart_disease.txt');  %%% Input data sets:  m=4, 2 classes,
%%%303 instances, 164 for class 1 and 139 for class 2. have 13 features (7
%%%categoric and 6 numeric). Maximum have 4 different categorical values. The last column is class label. 
x = xx(:,1:13); %%% samples
true_class = xx(:,14);  %%% class label
true_class = true_class';
m = 4;  %%% the maximum number of different values can be chosen by each atrribute  
x_categoric = [x(:, 2:3), x(:, 6:7), x(:, 9), x(:, 11), x(:, 13)]; %%% categorical part of the data
x_numeric = [x(:, 1), x(:, 4:5), x(:, 8), x(:, 10), x(:, 12)]; %%%% numerical part of the data

[n, d] = size(x);  %%% n: the number of samples;   d: the number of attributes(dimensionality)
[~, d_c] = size(x_categoric);
[~, d_n] = size(x_numeric);

%% Normalization to the numerical part %%%%%%%%%%%%%%%%%
for i=1:d_n
    x_numeric(:,i) = (x_numeric(:,i)-mean(x_numeric(:,i)))./std(x_numeric(:,i));  
end
%%%%% For Statlog Heart, Heart Disease, and Credit Approval data sets in the
%%%%% paper, it is better to normalize the numerical part first.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Implementation of OCIL algorithm %%%%%
totall_times = 50; %%% Execute 50 times to get the average performance
number_record = zeros(1, totall_times); % Record of obtained cluster number

for Times = 1:totall_times

    k = 3;  %%% initialized number of clusters
    cluster_label = zeros(1, n); %%% cluster label of each object
    cluster_count = ones(1, k); %%% number of objects in each cluster
    weight_cluster = ones(1, k); %%% weight of each cluster
    win_count = ones(1, k); %%% winning times of each cluster
    
    rho = 0.005;
    learning_rate = rho*k/n; %%% set the learning rate

    %% calculate the weight of each attribute (feature)%%%%%%%%%%%%%%%%
    sample_frequency = zeros(m, d_c);
    for i = 1:n
        for j = 1:d_c
            sample_frequency(x_categoric(i, j), j) = sample_frequency(x_categoric(i, j), j) + 1;
        end
    end

    num_of_diffvalue = zeros(1, d_c);
    for j = 1:d_c
        for i = 1:m
            if sample_frequency(i, j) ~= 0
                num_of_diffvalue(j) = num_of_diffvalue(j) + 1;
            end
        end
    end

    value_information = zeros(m, d_c);
    for i = 1:m
        for j = 1:d_c
            if sample_frequency(i, j) == 0
                value_information(i, j) = 0;
            else
                value_information(i, j) = -(sample_frequency(i, j)/n)*(log2(sample_frequency(i, j)/n));
            end
        end
    end
    attribute_entropy = sum(value_information)./num_of_diffvalue;
    feature_weight_categoric = (attribute_entropy./sum(attribute_entropy))*(d_c/(d_c+1)); %%% weight of each categorical attribute
    feature_weight_numeric = 1/(d_c + 1);
    feature_weight = [feature_weight_categoric, feature_weight_numeric];
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    %% Randomly initialize the k seed points (one for each class)
    miu_cate = zeros(k, d_c);  
    miu_numer = zeros(k, d_n);
    ran = randperm(n);
    for i = 1:k
        miu_cate(i,:) = x_categoric(ran(i), :);  
        miu_numer(i,:) = x_numeric(ran(i), :);
    end
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    numeric_sum = miu_numer;
    frequency_in_cluster = zeros(d_c, m, k);  %%%%count the frequency of each attribute value amoug the initial classes
    for c = 1:k
        for j = 1:d_c
            frequency_in_cluster(j,miu_cate(c,j),c) = frequency_in_cluster(j,miu_cate(c,j),c)+1;
        end
    end

    %% Penalized competitive learning %%%%%%%%%%
    max_epoch = 250; %% stopping criterion: maximum learning epochs 
    for t = 1 : max_epoch
        for i = 1 : n
        
            categoric_proba = zeros(d_c, k);
            for j = 1:d_c
                for c = 1:k
                    categoric_proba(j,c) = frequency_in_cluster(j, x_categoric(i, j), c)/cluster_count(c);
                end
            end
        
            average_numeric = zeros(k, d_n);
            for j = 1:k
                average_numeric(j, :) = numeric_sum(j, :)./cluster_count(j);
            end
            numeric_proba = zeros(1, k);
            for j = 1:k
                differen = x_numeric(i, :) - average_numeric(j, :);
                numeric_proba(j) = exp(-0.5*(differen*differen'));
            end
            numeric_proba = numeric_proba./sum(numeric_proba);
        
            attribute_proba = [categoric_proba; numeric_proba];
            class_proba = (attribute_proba')*(feature_weight');   %%%% class_proba is a column vector
            class_proba = class_proba'; %%% object-cluster similarity
        
            regulation = win_count ./ sum(win_count); %%% winning frequency
            weighted_class_proba = weight_cluster.*class_proba;
            weighted_dissim = regulation .* (1 - weighted_class_proba);
        
            [~, class_index] = sort(weighted_dissim);
            class_label = class_index(1);  %%% the winning seed point
            penaly_index = class_index(2); %%% the nearest rival to be penalized
            cluster_label(i) = class_label;
            for c = 1:k
                if c == class_label
                    numeric_sum(class_label, :) = numeric_sum(class_label, :) + x_numeric(i, :); 
                    for j = 1:d_c
                        frequency_in_cluster(j, x_categoric(i,j), class_label) = frequency_in_cluster(j, x_categoric(i,j), class_label)+1;
                    end
                    weight_cluster(c) = weight_cluster(c) + 1 * learning_rate; %%% reward the winner
                
                else if c == penaly_index
                        weight_cluster(c) = max(0, (weight_cluster(c) - learning_rate * (class_proba(c)))); %%% penalize the rival
                    end
                end
            end
            cluster_count(class_label) = cluster_count(class_label) + 1;
            win_count(class_label) = win_count(class_label) + 1;
        end
    end

    %% Analyze the result: calculate the obtained cluster number
    cluster_count = zeros(1, k);
    for i = 1:n
        cluster_count(cluster_label(i)) = cluster_count(cluster_label(i)) + 1;
    end
    cluster_number = 0;
    Threshold = 0.01 * n;
    %%% If the number of members in a cluster is less than one percent of 
    %%% the sample size, we can judge that this cluster has been eliminated.      
    for i = 1:k
        if cluster_count(i) >= Threshold
            cluster_number = cluster_number + 1;
        end
    end
    number_record(Times) = cluster_number; %%% record the obtained number of clusters

end
%% Summarize the results
number_mean = mean(number_record)
number_std = std(number_record)

